/*--------------------------------------------------------------------
 * $Id: mie_lex.l,v 1.24 2009/02/19 12:35:29 bernhard Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

%{

#include <getopt.h>
#include <stdlib.h>
#include <string.h>
#include "./mie.h"


int mie_check (input_struct input);

float  yytext2float(char* yytext, int which_token, int max_token);
float* yytext2floats(char* yytext, int nary);
int    yytext2nfloats(char* yytext);
int    yytext2int(char* yytext, int which_token, int max_token);
int*   yytext2integers(char* yytext, int nary);
int    yytext2nintegers(char* yytext);
char*  yytext2string(char* yytext, int which_token, int max_token);
int    yytext2ntokens(char* yytext);
int    yytextCstring(char *dest, char* yytext, int which_token, int max_token);


char* s, *t;
float* tmp;
int nary;
int ierror=0;
int line_number;
input_struct Input;
output_struct Output;
int ntokens=0;
int i=0;
int tmpint;

%}

%option noyywrap
%option nounput

DIGIT    [0-9]
SPACE    [ \t]
CCHAR    [#]
COMMENT  ({CCHAR}.*)
INTEGER  {DIGIT}+
SPACE_INTEGER {SPACE}+{INTEGER}
INTEGERS {SPACE_INTEGER}+ 
EXPON   ([DdEe][+-]?{DIGIT}+)
FLOAT    ((({DIGIT}+"."{DIGIT}*)|({DIGIT}+"."{DIGIT}*{EXPON}))|({INTEGER}))
SPACE_FLOAT {SPACE}+{FLOAT}
FLOATS   {SPACE_FLOAT}+ 
SIGN     [-+]
SIGNED_FLOAT {SIGN}?{FLOAT}
SPACE_SIGNED_FLOAT {SPACE}+{SIGNED_FLOAT}
SIGNED_FLOATS {SPACE_SIGNED_FLOAT}+
SIGNED_INTEGER {SIGN}?{INTEGER}
WORD     [_\.\/A-Za-z0-9][_\.\/A-Za-z0-9\-]*
SPACE_WORD {SPACE}+{WORD}
WORDS {SPACE_WORD}*

ZERO_OR_TWO_SIGNED_FLOATS ({SPACE}?{SIGNED_FLOAT}+{SPACE}+{SIGNED_FLOAT})?
ZERO_OR_TWO_OR_WORD {ZERO_OR_TWO_SIGNED_FLOATS}|{WORD}
FLOATS_OR_WORD ({FLOATS}|{SPACE_WORD})

%%

  /* These are the allowed input lines in the mie input file 
     The input line specifications must come before the eating of
     blanks, otherwise things get messed up 
     Obviously this whole lot could have been handled more properly
     using bison etc., but than rather few would know what was going 
     on??? 
   */

aerosol_type{SPACE}+{WORD}          {s =  yytext2string(yytext,2,2);
                                     Input.medium = AEROSOL;
                                     if (Input.distribution == DIST_NONE)
                                         Input.distribution = DIST_AER;
                                     if  (strcmp(s,"inso")==0) {Input.aerosol_type = TYPE_INSO;}
				     else if (strcmp(s,"waso")==0) {Input.aerosol_type = TYPE_WASO;}
				     else if (strcmp(s,"soot") ==0) {Input.aerosol_type = TYPE_SOOT;}
				     else if (strcmp(s,"ssam") ==0) {Input.aerosol_type = TYPE_SSAM;}
                                     else if (strcmp(s,"sscm") ==0) {Input.aerosol_type = TYPE_SSCM;}
                                     else if (strcmp(s,"minm") ==0) {Input.aerosol_type = TYPE_MINM;}
                                     else if (strcmp(s,"miam") ==0) {Input.aerosol_type = TYPE_MIAM;}
                                     else if (strcmp(s,"micm") ==0) {Input.aerosol_type = TYPE_MICM;}
                                     else if (strcmp(s,"mitr") ==0) {Input.aerosol_type = TYPE_MITR;}
                                     else if (strcmp(s,"suso") ==0) {Input.aerosol_type = TYPE_SUSO;}
                                     else {fprintf(stderr,"Unknown 'aerosol type %s' on line %d\n", s, line_number+1);
				       ierror++;}
                                     free(s);
                                     }

data_files_path{SPACE}+{WORD}       {yytextCstring (Input.data_files_path, yytext, 2, 2); 
                                     /* append slash */
                                     if (Input.data_files_path[strlen(Input.data_files_path)-1] != '/')
				       strcat (Input.data_files_path, "/");
                                    }

distribution{SPACE}+{WORD}{SPACE}+{FLOAT} {ntokens = yytext2ntokens(yytext);
                                     if (ntokens != 2) {
                                       fprintf(stderr, "Wrong number of arguments for 'distribution %s' on line %d\n", t, line_number+1);
				       ierror++;}
                                     s =  yytext2string(yytext,2,ntokens+1);
				     if      (strncmp(s,"none", 4)     ==0||strncmp(s,"NONE", 4)     ==0) {Input.distribution = DIST_NONE;}
				     else if (strncmp(s,"gamma",5)     ==0||strncmp(s,"GAMMA",5)     ==0) {Input.distribution = DIST_GAMMA;}
				     else if (strncmp(s,"lognormal",9) ==0||strncmp(s,"LOGNORMAL",9) ==0) {Input.distribution = DIST_LOGNORMAL;}
				     else if (strncmp(s,"file",4)      ==0||strncmp(s,"FILE",4)      ==0) {Input.distribution = DIST_FILE;}
                                     else {fprintf(stderr,"Unknown 'distribution %s' on line %d\n", s, line_number+1);
				       ierror++;}
                                     free(s);
                                     if (Input.distribution == DIST_GAMMA || Input.distribution == DIST_LOGNORMAL)
                                       Input.alpha = yytext2float(yytext, 3, ntokens+1);
                                     else 
                                       Input.sd_filename = yytext2string(yytext,3,ntokens+1);
                                    }
mass_density{SPACE}+{SIGNED_FLOAT}  {Input.rho_user = yytext2float(yytext,2,2); }
mie_program{SPACE}+{WORD}           {s =  yytext2string(yytext,2,2);
				     if      (strncmp("MIEV0",s,5)==0) {Input.program = MIEV0;}
				     else if (strncmp(s,"BH",2)   ==0) {Input.program = BHMIE;}
				     else {fprintf(stderr,"Unknown 'mie_program %s' on line %d\n", t, line_number+1);
				       ierror++;}
                                     free(s);
                                    }
mimcut{SPACE}+{SIGNED_FLOAT}        {Input.mie.mimcut = yytext2float(yytext,2,2); }
nmom{SPACE}+{INTEGER}               {Input.mie.nmom   = yytext2int(yytext,2,2); 
                                     Input.mie.momdim = Input.mie.nmom+1;}
nstokes{SPACE}+{INTEGER}            {Input.nstokes   = yytext2int(yytext,2,2); } 
                                     

output_user{WORDS}                  {Input.n_output_user = yytext2ntokens(yytext);
				     Input.output_user = calloc (Input.n_output_user, sizeof (int));
				     for (i=2;i<=Input.n_output_user+1; i++) {
				       t = yytext2string(yytext, i, Input.n_output_user+1);  
                                       if      (strcmp("none",t)      ==0)  {Input.output_user[i-2] = OUTPUT_USER_NONE;}
				       else if (strcmp("lambda",t)==0 || strcmp("wavelength",t)==0)
                                                                            {Input.output_user[i-2] = OUTPUT_USER_WAVE;}
				       else if (strcmp("wavenumber",t) ==0) {Input.output_user[i-2] = OUTPUT_USER_WAVENUMBER;}
				       else if (strcmp("refrac_real",t)==0) {Input.output_user[i-2] = OUTPUT_USER_REFRAC_REAL;}
				       else if (strcmp("refrac_imag",t)==0) {Input.output_user[i-2] = OUTPUT_USER_REFRAC_IMAG;}
                                       else if (strcmp("r_eff"      ,t)==0) {Input.output_user[i-2] = OUTPUT_USER_R_EFF;}
				       else if (strcmp("qext"       ,t)==0) {Input.output_user[i-2] = OUTPUT_USER_QEXT;}
                                       else if (strcmp("qsca"       ,t)==0) {Input.output_user[i-2] = OUTPUT_USER_QSCA;}
				       else if (strcmp("omega"      ,t)==0) {Input.output_user[i-2] = OUTPUT_USER_OMEGA;}
				       else if (strcmp("gg"         ,t)==0) {Input.output_user[i-2] = OUTPUT_USER_GG;}
				       else if (strcmp("spike"      ,t)==0) {Input.output_user[i-2] = OUTPUT_USER_SPIKE;}
				       else if (strcmp("pmom"       ,t)==0) {Input.output_user[i-2] = OUTPUT_USER_PMOM;}
				       else if (strcmp("sforw"      ,t)==0) {Input.output_user[i-2] = OUTPUT_USER_SFORW;}
				       else if (strcmp("sback"      ,t)==0) {Input.output_user[i-2] = OUTPUT_USER_SBACK;}
				       else if (strcmp("cloudprp"   ,t)==0) {Input.output_format = OUTPUT_FORMAT_CLOUDPROP;}
				       else if (strcmp("aerosolprp" ,t)==0) {Input.output_format = OUTPUT_FORMAT_AERPROP;}
				       else if (strcmp("netcdf"     ,t)==0) {Input.output_format = OUTPUT_FORMAT_NETCDF;}
				       else {fprintf(stderr,"Unknown 'output_user %s' on line %d\n", t, line_number+1);
				       ierror++;}
				       free(t);
				     }
                                    }
r_eff{SIGNED_FLOATS}|r_mean{SIGNED_FLOATS} {ntokens = yytext2ntokens(yytext);
                                     Input.r_eff_min    = yytext2float(yytext, 2, ntokens+1);
                                     if (ntokens>1)
                                       Input.r_eff_max  = yytext2float(yytext, 3, ntokens+1);
                                     if (ntokens>2)
                                       Input.r_eff_step = yytext2float(yytext, 4, ntokens+1);
                                    }
refrac{SPACE}+{WORD}{SPACE}?{ZERO_OR_TWO_OR_WORD}  {
                                       s = (char *) calloc (strlen(yytext)+1,sizeof(char));
                                       strcpy (s, yytext); 
				       t =  yytext2string(s,2,4);
				       if (strncmp("ice",t,3)==0) 
					 Input.medium = ICE;
				       else  {
					 if (strncmp("water",t,5)==0) 
					   Input.medium = WATER;
					 else  {
					   if (strncmp("user",t,4)==0) {
					     Input.medium = USER;
					     
					     strcpy (s, yytext); 
					     nary = yytext2nfloats(s);
					     
					     strcpy (s, yytext);
					     tmp = yytext2floats(s, nary);
					     
					     Input.crefin.re = tmp[1];
					     Input.crefin.im = tmp[2];
					   }
					   else {
					     if (strncmp("file",t,4)==0) {
					       Input.medium = USER;
					       Input.ri_filename = yytext2string(yytext,3,3); 
					     }
                                             else {
                                               fprintf(stderr,"Unknown command on line %d\n", line_number+1);
                                               fprintf(stderr,"%s\n", yytext);
                                               ierror++;
                                             }
					   }
					 }
				       }
				       free(s);}
size_distribution_file{SPACE}+{WORD} {Input.sd_filename = yytext2string(yytext,2,2);
                                      Input.distribution = DIST_FILE; }
temperature{SPACE}+{SIGNED_FLOAT}    {Input.temperature   = yytext2float(yytext,2,2); }
wavelength{FLOATS_OR_WORD}|wvn{FLOATS_OR_WORD} {
                                       ntokens = yytext2ntokens(yytext);
                                       if (ntokens == 1) {
                                         s = yytext2string(yytext,2,ntokens+1);
				         if      (strlen(s)==8 && strncasecmp("kato2.96",s,8)    ==0) {Input.ck_scheme = CK_KATO2_96;}
				         else if (strlen(s)==5 && strncasecmp("kato2",s,5)       ==0) {Input.ck_scheme = CK_KATO2;}
				         else if (strlen(s)==4 && strncasecmp("kato",s,4)        ==0) {Input.ck_scheme = CK_KATO;}
				         else if (                strncasecmp("fu",s,2)          ==0) {Input.ck_scheme = CK_FU;}
				         else if (                strncasecmp("avhrr_kratz",s,11)==0) {Input.ck_scheme = CK_AVHRR_KRATZ;}
				         else if (                strncasecmp("sbdart",s,6)      ==0 || 
                                                                  strncasecmp("lowtran",s,7)     ==0) {Input.ck_scheme = CK_LOWTRAN;}
				         else                                                         {Input.ck_scheme = CK_GENERIC;
                                           Input.filename_ck_generic = (char *) calloc (strlen(s)+1,sizeof(char));
				           strcpy (Input.filename_ck_generic, s);}
				         free(s);
                                       }
                                       else if (ntokens == 2) {
				         Input.wl.wl_start  = yytext2float(yytext,2,ntokens+1); 
                                         Input.wl.wl_end    = yytext2float(yytext,3,ntokens+1);
                                       }
                                       else {
                                         fprintf(stderr,"Unknown command on line %d\n", line_number+1);
                                         fprintf(stderr,"%s\n", yytext);
                                         ierror++;
                                       } 
				     }
wavelength_step{SPACE}+{SIGNED_FLOAT}|wvn_step{SPACE}+{SIGNED_FLOAT}       {Input.wl.wl_step   = yytext2float(yytext,2,2); }
wavelength_index{SPACE_INTEGER}+{SPACE_INTEGER}|wvn_index{SPACE_INTEGER}+{SPACE_INTEGER}  {s = (char *) calloc (strlen(yytext)+1,sizeof(char));
                                     strcpy (s, yytext);
				     Input.lambda_i_start = yytext2int(s,2,3); 
                                     Input.lambda_i_end   = yytext2int(yytext,3,3);
				     if (Input.lambda_i_start > Input.lambda_i_end) {
				       tmpint = Input.lambda_i_start;
				       Input.lambda_i_start = Input.lambda_i_end;
				       Input.lambda_i_end   = tmpint;
				     }
                                     free(s);}


{SPACE}+       ;         /* eat up blanks */ 

{COMMENT}*     ;         /* eat up comments */

\n                        { ++line_number;};

{WORD}                    {fprintf(stderr,"Unknown command on line %d\n", line_number+1);
                           fprintf(stderr,"%s\n", yytext);
                           ierror++;}

<<EOF>> { 
  if (ierror > 0) 
    exit(1);
  
  yyterminate();
}      

%%

/*************************************************************/
/* print usage information                                   */
/*************************************************************/

static void usage()
{
  fprintf (stderr, "\nCalculates scattering and absorption properties using Mie theory.\n");
  fprintf (stderr, "See libRadtran documentation for more information.\n\n");
  fprintf (stderr, "No error checking yet, so be careful.\n\n");

  fprintf (stderr, "Usage: mie [-h] < input_file > Output_file\n");
  fprintf (stderr, " -h    Print this message.\n");  
  fprintf (stderr, "\n");  
}

int main(int argc, char **argv) 
{
  
  /*  @12c@                

@code{mie} performes Mie scattering calculations for a specified
wavelength interval. It reads input from standard input, 
and outputs to standard output. It is normally invoked in the following
way: 

@example
mie < input_file > output_file
@end example

@noindent The format of the input and output files are described
below. Several realistic examples of input files are subsequently 
given.

@strong{Warning:} Please note the error checking on input variables 
is very scarce at the moment. Hence, if you provide erroneous input, 
the outcome is unpredictable.
  @c12@ */

  /*  @12_01c@                
@cindex input file, mie
@cindex mie, input file

The mie input file consists of single line entries, each making
up a complete input to the mie program. First on the line comes
the parameter name, followed by one or more parameter values. The 
parameter name and the parameter values are seperated by white space.

Filenames are entered without any surrounding single or double quotes.

Comments are introduced by a @code{#}. Blank lines are ignored.

The various input parameters are described in detail below.
@vtable @code
@item aerosol_type
With this option Mie calculations are performed for the specified aerosol type. The 
aerosol properties (refractive index, size distribution, density, humidity) are taken 
from the OPAC database (see Hess et.al., 1998). 
Possible types are
@vtable @code
@item inso 
Water insoluble aerosol consists mostly of soil particles with a certain amount of organic material. 
@item waso
Water soluble aerosol originates from gas to particle conversion and consists of various types of sulfates, nitrates, and other, also organic water-soluble substances.
@item soot
Soot is absorbing black carbon, which is not soluble in water. In reality soot particles have a chain-like character, which of course is not accounted for in Mie calculations of optical properties. The optical properties are calculated assuming many very small spherical particles.  
@item ssam 
Sea salt particles consist of the various kinds of salt contained in seawater. The different modes are given to allow for a different wind-speed-dependant increase of particle number for particles of different size.
This aerosol type represents the accumulation mode.
@item sscm 
Sea salt particles (coarse mode).
@item minm  
Mineral aerosol or desert dust is produced in arid regions. It consists of a mixture of quartz and clay minerals and is modeled with three modes to allow to consider increasing relative amount of large particles for increased turbidity.
This aerosol type represents the nucleation mode.
@item miam
Mineral aerosol (accumulation mode). 
@item micm
Mineral aerosol (coarse mode). 
@item mitr  
Mineral transported is used to describe desert dust that is transported over long distances with a reduced amount of large particles.
@item suso
The sulfate component is used to describe the amount of sulfate found in the Antarctic aerosol. This component is not suited to describe antropogenic sulfate aerosols that are included in the water-soluble component.  
@end vtable  
@item distribution
If specified the effecive radius is converted into a size distribution of droplets.
@example
distribution distribution_type distribution_parameter
@end example
where distribution type is one of the two following:
@vtable @code
@item GAMMA
The  Gamma distribuition of cloud droplet sizes is 
@example
n(r) = a r^alpha exp(-b r),
@end example
where alpha is the distribution paramter given as second argument. (@code{a} and @code{b}
are determined automatically.) The effective radius of the distribution is r_eff = (alpha+3)/b.
A typical value for water clouds is alpha=7. For ice clouds a typical value is alpha=1.
A large value of alpha gives close to a monodisperse distribution.
@item LOGNORMAL
The lognormal distribution of cloud droplet sizes is 
@example
n(r) = a/r exp (-[ln(r/r0)]^2 / [2 sigma^2]),
@end example
where r0 is the logarithmic mode of the distribution (calculated automatically) and sigma is 
the standard deviation, which is given by second argument.
The effective radius of the distribution is r_eff = r0 exp(2.5 sigma^2).
A common value for water clouds is sigma=0.35.
@end vtable
@item mass_density
Specifies the mass density of the medium.
Useful in combinition with @code{refrac user} and @code{output_user cloudprop},
as the format of @code{cloudprop} specifies the extinction coefficient per 
mass and not per volumn as usual in this mie program.
@item mie_program
Specify which Mie program to use:
@vtable @code
@item BH
The Mie scattering program by Bohren and Hoffmann, @*
ftp://astro.princeton.edu/draine/scat/bhmie/
@item MIEV0
The Mie scattering program by W. Wiscombe. For documentation see 
src_f/MIEV.doc and the NCAR Mie report at @*
ftp://climate.gsfc.nasa.gov/pub/wiscombe/Single_Scatt/.
@end vtable
@item mimcut
(positive) value below which imaginary refractive
index is regarded as zero (computation proceeds
faster for zero imaginary index). Only used by @code{mie_program} @code{MIEV0}.
@item nmom 
Number of moments of the phase function to be calculated (default: 0). 
Only possible with @code{mie_program} @code{MIEV0}.
@item nstokes 
Number of Stokes parameters (default: 1). For @code{nstokes=1} the
Legendre polynomials of the phase function will be calculated. To
calculate all phase matrix elements required for polarized radiative
transfer, set @code{nstokes=4}.      
@item output_user
The mie output is one line of output quantities to standard output (stdout) for each 
wavelength and each particle radius.
With this option the user may specify the columns desired for output:
@example
output_user output_1 output_2 ... output_n
@end example
where output_i is one of following arguments:
@vtable @code
@item lambda
Wavelength in nm.
@item wavenumber
Wave number in cm-1.
@item r_eff
particle radius in micro meter.
@item refrac_real
The real part of the refractive index.
@itemx refrac_imag
The imaginary part of the refractive index.
@item qext
The extinction efficiency factor, if @code{r_eff} was specified, or
the extinction coefficient [km-1] per unit concentration [cm3/m3],
if a @code{size_distribution_file} was specified. If the medium is 
liquid water, 1 cm3/m3 equals a liquid water content of 1g/m3 because 
the density of water is close to 1 g/cm3. For ice and other substances,
the density has to be considered (0.917 g/cm3 for ice at 273K).
@item qsca
The scattering efficiency factor, if @code{r_eff} was specified, or
the extinction coefficient [km-1] per unit concentration [cm3/m3],
if a @code{size_distribution_file} was specified.
@item omega
The single scattering albedo.
@item gg
The asymmetry parameter.
@item sforw
(Complex) forward-scattering amplitude S1 at 0 degrees.
@item sback
(Complex) back-scattering amplitude S1 at 180 degrees. 
@item spike
To quote from Wiscombe's @file{MIEV0.doc}:
     @quotation
(REAL) magnitude of the smallest denominator of
            either Mie coefficient (a-sub-n or b-sub-n),
            taken over all terms in the Mie series past
            N = size parameter XX.  Values of SPIKE below
            about 0.3 signify a ripple spike, since these
            spikes are produced by abnormally small denominators
            in the Mie coefficients (normal denominators are of
            order unity or higher).  Defaults to 1.0 when not
            on a spike.  Does not identify all resonances
            (we are still working on that).
     @end quotation
Meaningless if a @code{size_distribution_file} was specified.
@item pmom
The nmom+1 moments (from 0 to nmom, see option @code{nmom}) of the phase function. 
The phase function 
@iftex
 @tex
 $p(\mu)$ is @*
  $$ p (\mu) = \sum_{m=0}^{\infty} (2m+1) \cdot k_m \cdot P_m (\mu) $$
  where $k_m$ is the m'th moment and $P_m (\mu)$ is the m'th Legendre polynomial.
 @end tex
@end iftex
@ifinfo
 p(mu) is @*
 p (mu) = sum (m=0 to infinity) (2m+1) * k(m) * Pm (mu) @*
 where k(m) is the m'th moment and Pm (mu) is the m'th Legendre polynomial.
@end ifinfo
@item cloudprp
This is a special option which, if specified, must be the only option of @code{output_user},
as @code{cloudprp} specifies a whole format of the output. In particular this option is 
useful when a correlated-k wavelength grid is specified with @code{wavelength}. 
If specified, the output will be written in a format, which can be directly used by 
@code{libRadtran}. See also @code{output netcdf}.
See @code{libRadtran}-manual @code{ic_properties} and @code{wc_properties} and 
there the items @code{mie} and @code{filename}.
@item aerosolprp
This option is similar to the @code{cloudprp} option. The only difference is that 
the effective radius dimension is replaced by humidity values of the aerosol. 
@item netcdf 
This option writes the same output as @code{cloudprp} and @code{aerosolprp} directly in netCDF 
format. The outputfiles can be used by @code{uvspec} using the options @code{ic_properties} 
and @code{wc_properties}.
@end vtable
The default output is: 
@example
lambda refrac_real refrac_imag qext omega gg spike pmom
@end example
@item r_eff
The radius [micron] of the particle to calculate single scattering properties of.
Used together with the wavelength information to calculate the 
Mie size parameter.
@example
r_eff radius
@end example
The user can optionally specify a 2nd and 3rd argument to make a loop over several radii:
@example
r_eff radius_min radius_max radius_step
@end example
First calculations is done with radius_min, which will be increased by radius_step
until radius_max is reached.
@item refrac
Specify which refractive index to use. The following options are implemented: 
@vtable @code
@item ice
The complex refractive index is taken from the REFICE function of W. Wiscombe. 
@item water
The complex refractive index is taken from the REFWAT function of W. Wiscombe. 
@item user <re> <im>
A user defined refractive index. re and im are the real and imaginary parts
(both positive numbers).
@item file <filename>
Read refractive index from a three-column file containing wavelength [nm],
and the real and imaginary parts of the refractive index (both positive numbers).
The Mie calculation is done for each wavelength defined here.
@end vtable
@item size_distribution_file
Specify a two column file, r [micron], n(r), which describes a size distribution 
of droplets. The Mie calculation is repeated for each value of r found in the 
size distribution file, and the final result is a weighted average of these 
values. The user himself therefore has to choose a set of r's suited for 
his specific purpose.
@item temperature
Ambient temperature, used to calculate the refractive indices of water and ice.
Temperature dependence is only considered above 10 micron (water) and 167 micron (ice), 
respectively. Default: 300K.
@item wavelength
Sets the wavelength range, in nm. 
@example
wavelength lambda_min lambda_max
@end example
The wavelength step is specified by @code{wavelength_step}.

For unregular wavelength grid it is also possible to specify a file, where the 
wavelengthgrid is stored.
@example
wavelength wvl_filename
@end example
where @code{wvl_filename} is the path and name of the file, which contains 
the wavelength grid. It is expected that the wavelength values in nm are stored 
in the second column.
For the @code{correlated_k} schemes implemented in @code{uvspec} you can use following abbrevations
instead of a filename (in this case @code{data_files_path} must be specified also): 
@code{kato}, @code{kato2}, @code{kato2.96}, @code{fu}, and @code{avhrr_kratz}.

This option is ignored if @code{refrac file} is specified.
@item wavelength_step
The wavelength step, in nm. Ignored if @code{refrac file} is specified.
@item wavelength_index
Set the wavelengths to be selected. This might be the normal wavelength grid defined by 
@code{wavelength} and @code{wavelength_step} or a @code{correlated_k} wavelength grid.
E.g., in combination with
@code{wavelength AVHRR_KRATZ}, @code{wavelength_index 15 15} will select wavelength
index 15 which corresponds to channel 4, or @code{wavelength_index 10 14} will select 
those bands required for channel 3. Indices start from 1.

@end vtable
  @c12_01@ */

  /*  @12_02c@                
@cindex output, mie
@cindex uvspec, mie
standard output (stdout) of the mie program is one line for each 
wavelength and each effective radius.
The format of the output line is
@example
lambda refrac_real refrac_imag qext omega gg spike pmom
@end example
The keywords here are the same as in input option @code{output_user}.

  @c12_02@ */

  /*  @12_03c@                

An example of a complete input file is

@example
@include ../examples/MIE_1.INP
@end example


      @c12_03@   */

  /***************** End of documentation *********************/


  /* prototype of main function */
  int mie(input_struct input, output_struct *output);

  int c;
  /* int iu, iv, j, lev; */
  int status=0;
  /*  Input_struct Input;
  Output_struct Output;*/

  /* Initialization */

  Output.print.phi    = 1; /* Default print radiances */


  Input.medium          =  WATER;
  Input.temperature     =  300.0;
  Input.rho_user        = -999.0;
  Input.r_eff_min       = -999.0;
  Input.r_eff_max       = -999.0;
  Input.distribution    =  DIST_NONE;
  Input.alpha           = -999.0;
  Input.sd_filename     =  "";
  Input.ri_filename     =  "";
  Input.program         =  MIEV0;
  Input.nstokes         = 1;
  
  Input.wl.wl_start     = 0.0;
  Input.wl.wl_end       = 0.0;
  Input.mie.anyang      = 1;
  Input.mie.ipolzn      = 0;
  Input.mie.momdim      = 0;
  Input.mie.nmom        = 0; 
  Input.mie.numang      = 0;
  Input.mie.mimcut      = 1.0e-08;
  Input.mie.perfct      = 0;
  Input.mie.prnt[0]     = 0;
  Input.mie.prnt[1]     = 0;
  Input.mie.xmu         = (float *) calloc (Input.mie.numang+1, sizeof(float));
  Input.mie.s1          = (mie_complex *) calloc (Input.mie.numang+1, sizeof(mie_complex));
  Input.mie.s2          = (mie_complex *) calloc (Input.mie.numang+1, sizeof(mie_complex));
  Input.n_output_user   = 0;
  Input.ck_scheme       = CK_NONE;
  Input.lambda_i_start  = -999;
  Input.lambda_i_end    = -999;
  Input.data_files_path = (char *) calloc (FILENAME_MAX,sizeof(char));
  
  while ((c=getopt (argc, argv, "h")) != EOF)  {
    switch(c)  {
    case 'h': 
      usage();
      return (-1);
      break;
    default:
      usage();
      return (-1);
    }
  }

  /* Read the input file */

  yylex();

  /* initialisation dependent on input data */

  if (Input.r_eff_max == -999.)
    Input.r_eff_max = Input.r_eff_min;
  
  switch (Input.medium) {
    case WATER:
      Output.rho_medium = RHO_WATER;
      break;
    case ICE:
      Output.rho_medium = RHO_ICE;
      break;
    case USER:
      if (Input.rho_user > 0.0)
        Output.rho_medium = Input.rho_user;
      else 
        Output.rho_medium = 1.0;
      break;
    case AEROSOL: 
       fprintf(stderr, "Mie calculation for aerosol. Size distributions, refractive indices\n");
       fprintf(stderr, "and the aerosol density are taken from the OPAC database.\n");
       break; 
    default:
      fprintf (stderr, "Error, unknown medium %d\n", Input.medium);
      break; 
  }

 
  /* checks */

  status = mie_check(Input);
  if (status!=0) {
    fprintf (stderr, "%d error(s) in mie input-file, aborting\n", status);
    return status;
  }

  /* Do the Mie calculation */

  status = mie(Input, &Output);
  if (status != 0.0) {
    fprintf (stderr, "Error during execution of mie\n");
    return status;
  }

  return status;

}

float yytext2float(char* yytext, int which_token, int max_token) {
  double dnumber;
  char token[] = " ";
  char *p, *s=NULL;
  int count;
  char *yytext2=NULL;

  yytext2 = (char *) calloc (strlen(yytext)+1,sizeof(char));
  strcpy (yytext2, yytext);

  p = strtok(yytext2,token);
  count = 1;
  while ((p = strtok(NULL,token)) != NULL) {
    count++;
    if (count == which_token)
      s = p;
  }
  if ((count != which_token) && (count>max_token)) {
    fprintf(stderr, "Wrong number of arguments (%d) on line %d\n", count, line_number+1);
  }
  dnumber = atof(s);
  /*   printf("dnumner %f\n",dnumber);*/
  free(yytext2);
  
  return (float)dnumber;
}

float* yytext2floats(char* yytext, int nary) {
  int i;
  char token[] = " ";
  char *p;
  char *yytext2=NULL;

  /* Fill up ary */
  float *ary    = (float *) calloc (nary, sizeof(float));

  yytext2 = (char *) calloc (strlen(yytext)+1,sizeof(char));
  strcpy (yytext2, yytext);

  p = strtok(yytext2,token);
  i = 0;
  while ((p = strtok(NULL,token)) != NULL)
    ary[i++] = atof(p);

  return ary;
}

int yytext2nfloats(char* yytext) {
  int nary;
  char token[] = " ";
  char *p;
  char *yytext2=NULL;

  yytext2 = (char *) calloc (strlen(yytext)+1,sizeof(char));
  strcpy (yytext2, yytext);

  /* Get number of ary elements */
  nary = 0;
  p = strtok(yytext2,token);
  while ((p = strtok(NULL,token)) != NULL) 
    nary++;

  return nary;
}

int* yytext2integers(char* yytext, int nary) {
  int i;
  char token[] = " ";
  char *p;
  char *yytext2=NULL;

  /* Fill up ary */
  int *ary    = (int *) calloc (nary, sizeof(int));

  yytext2 = (char *) calloc (strlen(yytext)+1,sizeof(char));
  strcpy (yytext2, yytext);

  p = strtok(yytext2,token);
  i = 0;
  while ((p = strtok(NULL,token)) != NULL)
    ary[i++] = atoi(p);

  return ary;
}

int yytext2nintegers(char* yytext) {
  int nary;
  char token[] = " ";
  char *p;
  char *yytext2=NULL;

  yytext2 = (char *) calloc (strlen(yytext)+1,sizeof(char));
  strcpy (yytext2, yytext);

  /* Get number of ary elements */
  nary = 0;
  p = strtok(yytext2,token);
  while ((p = strtok(NULL,token)) != NULL)
    nary++;

  return nary;
}

int yytext2int(char* yytext, int which_token, int max_token) {
  char token[] = " ";
  char *p, *s=NULL;
  int count;
  char *yytext2=NULL;

  yytext2 = (char *) calloc (strlen(yytext)+1,sizeof(char));
  strcpy (yytext2, yytext);

  p = strtok(yytext2,token);
  count = 1;
  while ((p = strtok(NULL,token)) != NULL) {
    count++;
    if (count == which_token)
      s = p;
  }

  if ((count != which_token) && (count>max_token))
    fprintf(stderr, "Wrong number of arguments on line %d\n", line_number+1);

  return atoi(s);
}

char* yytext2string(char* yytext, int which_token, int max_token) {
  char token[] = " ";
  char *p, *s=NULL, *string=NULL;
  int count;
  char *yytext2=NULL;

  yytext2 = (char *) calloc (strlen(yytext)+1,sizeof(char));
  strcpy (yytext2, yytext);

  p = strtok(yytext2,token);
  count = 1;
  while ((p = strtok(NULL,token)) != NULL) {
    count++;
    if (count == which_token)
      s = p;
  }
  if ((count != which_token) && (count>max_token)) {
    fprintf(stderr, "Wrong number of arguments (%d) on line %d\n", count, line_number+1);
  }
  string = (char *) calloc (strlen(s)+1, sizeof (char));
  strcpy (string, s);
   
  free(yytext2);   
  return string;
}

int yytext2ntokens(char* yytext) {
  int nary=0;
  char token[]=" ";
  char *p=NULL;
  char *yytext2=NULL;

  yytext2 = (char *) calloc (strlen(yytext)+1,sizeof(char));
  strcpy (yytext2, yytext);

  /* Get number of ary elements */
  nary = 0;
  p = strtok(yytext2,token);

  while ((p = strtok(NULL,token)) != NULL) 
    nary++;

  free(yytext2);
  return nary;
}

/* like yytext2string, but string is copied to a pre-allocated string dest */
int yytextCstring(char *dest, char* yytext, int which_token, int max_token) {
  char token[]=" ";
  char *p=NULL, *s=NULL, *tmpstring=NULL;
  int count=0;

  tmpstring = (char *) calloc (strlen(yytext)+1, sizeof (char));
  strcpy (tmpstring, yytext);

  p = strtok(tmpstring,token);
  count = 1;
  while ((p = strtok(NULL,token)) != NULL) {
    count++;
    if (count == which_token)
      s = p;
  }

  if ((count != which_token) && (count>max_token))
    fprintf(stderr, "Wrong number of arguments on line %d\n", line_number+1);

  strcpy (dest, s);

  free(tmpstring);

  return 0;
}




/**************************************************/
/* Check the Mie input structure for plausibility */
/**************************************************/

int mie_check (input_struct input)
{
  int status=0;
  status =0;

  if (!( input.nstokes==1 || input.nstokes==4)){
    fprintf(stderr, "Error: nstokes must be either 1 or 4 \n");
    status = -1;
   }

  if (input.nstokes >1 && input.program != MIEV0){
    fprintf(stderr, "Error: To calclate the phase matrix (*nstokes*=4) you must use the mie_program MIEV0.\n");
    status = -1;
   }
  
   
  if (input.mie.nmom >0 && input.program != MIEV0){
    fprintf(stderr, "Error: To calclate the Legendre polynomials you must use the mie_program MIEV0.\n");
    status = -1;
   } 

 if (Input.aerosol_type!=0 && Input.medium != AEROSOL){
     fprintf(stderr, "*** Warning: You have chosen an aerosol type and provided a refractive index.\n");
     fprintf(stderr, "*** The refractive index of the aerosol type will be used for the calculation,\n");
     fprintf(stderr, "*** not the one that you have provided. \n");
   }

   if (Input.aerosol_type!=0 && Input.distribution != DIST_AER){
        fprintf(stderr, "Error: You have chosen an aerosol type and provided a size distribution.\n");
        fprintf(stderr, "       This is not possible since the OPAC size distributions are \n");
        fprintf(stderr, "       taken automatically. Please remove the option *distribution*.\n");
        status = -1;
   }
    
   if (Input.medium != AEROSOL &&  Input.r_eff_min==-999){
       fprintf(stderr, "Error: You have to provide a particle size to perform a Mie calculation \n");
       fprintf(stderr, "       for clouds. Please specify *reff*.\n"); 
       status = -1; 
    }

   if (Input.medium != AEROSOL && Input.output_format == OUTPUT_FORMAT_AERPROP){
       fprintf(stderr, "Error: You have chosen *output_user* aerosolprp but you want to do a calculation \n");
       fprintf(stderr, "       for clouds. Please use cloudprp instead. \n");
       status = -1;
    }

    if (Input.medium == AEROSOL && Input.output_format == OUTPUT_FORMAT_CLOUDPROP){
       fprintf(stderr, "Error: You have chosen *output_user* cloudprp but you want to do a calculation \n");
       fprintf(stderr, "       for aerosols. Please use aerosolprp instead. \n");
       status = -1;
     }

    if (Input.output_format == OUTPUT_FORMAT_CLOUDPROP && Input.distribution == DIST_NONE){
       fprintf(stderr, "Error: You have chosen *output_user* cloudprp. For this option you have \n");
       fprintf(stderr, "       to specifiy a size distribution using the option *distribution*. \n");
       status = -1; 
     }

  return status;
}
